@@ -1,7 +1,20 @@
 use bevy::prelude::*;
 use bevy_egui::{egui, EguiContext, EguiPlugin, EguiSettings};
 
-const BEVY_TEXTURE_ID: u64 = 0;
+struct Images {
+    bevy_icon: Handle<Image>,
+    bevy_icon_inverted: Handle<Image>,
+}
+
+impl FromWorld for Images {
+    fn from_world(world: &mut World) -> Self {
+        let asset_server = world.get_resource_mut::<AssetServer>().unwrap();
+        Self {
+            bevy_icon: asset_server.load("icon.png"),
+            bevy_icon_inverted: asset_server.load("icon_inverted.png"),
+        }
+    }
+}
 
 /// This example demonstrates the following functionality and use-cases of bevy_egui:
 /// - rendering loaded assets;
@@ -14,7 +27,6 @@ fn main() {
         .init_resource::<UiState>()
         .add_plugins(DefaultPlugins)
         .add_plugin(EguiPlugin)
-        .add_startup_system(load_assets)
         .add_startup_system(configure_visuals)
         .add_system(update_ui_scale_factor)
         .add_system(ui_example)
@@ -27,16 +39,12 @@ struct UiState {
     value: f32,
     painting: Painting,
     inverted: bool,
-}
-
-fn load_assets(mut egui_context: ResMut<EguiContext>, assets: Res<AssetServer>) {
-    let texture_handle = assets.load("icon.png");
-    egui_context.set_egui_texture(BEVY_TEXTURE_ID, texture_handle);
+    egui_texture_handle: Option<egui::TextureHandle>,
 }
 
 fn configure_visuals(mut egui_ctx: ResMut<EguiContext>) {
     egui_ctx.ctx_mut().set_visuals(egui::Visuals {
-        window_corner_radius: 0.0,
+        window_rounding: 0.0.into(),
         ..Default::default()
     });
 }
@@ -64,12 +72,33 @@ fn update_ui_scale_factor(
 fn ui_example(
     mut egui_ctx: ResMut<EguiContext>,
     mut ui_state: ResMut<UiState>,
-    assets: Res<AssetServer>,
+    // You are not required to store Egui texture ids in systems. We store this one here just to
+    // demonstrate that rendering by using a texture id of a removed image is handled without
+    // making bevy_egui panic.
+    mut rendered_texture_id: Local<egui::TextureId>,
+    mut is_initialized: Local<bool>,
+    // If you need to access the ids from multiple systems, you can also initialize the `Images`
+    // resource while building the app and use `Res<Images>` instead.
+    images: Local<Images>,
 ) {
+    let egui_texture_handle = ui_state
+        .egui_texture_handle
+        .get_or_insert_with(|| {
+            egui_ctx
+                .ctx_mut()
+                .load_texture("example-image", egui::ColorImage::example())
+        })
+        .clone();
+
     let mut load = false;
     let mut remove = false;
     let mut invert = false;
 
+    if !*is_initialized {
+        *is_initialized = true;
+        *rendered_texture_id = egui_ctx.add_image(images.bevy_icon.clone_weak());
+    }
+
     egui::SidePanel::left("side_panel")
         .default_width(200.0)
         .show(egui_ctx.ctx_mut(), |ui| {
@@ -80,6 +109,11 @@ fn ui_example(
                 ui.text_edit_singleline(&mut ui_state.label);
             });
 
+            ui.add(egui::widgets::Image::new(
+                egui_texture_handle.id(),
+                egui_texture_handle.size_vec2(),
+            ));
+
             ui.add(egui::Slider::new(&mut ui_state.value, 0.0..=10.0).text("value"));
             if ui.button("Increment").clicked() {
                 ui_state.value += 1.0;
@@ -93,7 +127,7 @@ fn ui_example(
             });
 
             ui.add(egui::widgets::Image::new(
-                egui::TextureId::User(BEVY_TEXTURE_ID),
+                *rendered_texture_id,
                 [256.0, 256.0],
             ));
 
@@ -151,15 +185,16 @@ fn ui_example(
         ui_state.inverted = !ui_state.inverted;
     }
     if load || invert {
-        let texture_handle = if ui_state.inverted {
-            assets.load("icon_inverted.png")
+        // If an image is already added to the context, it'll return an existing texture id.
+        if ui_state.inverted {
+            *rendered_texture_id = egui_ctx.add_image(images.bevy_icon_inverted.clone_weak());
         } else {
-            assets.load("icon.png")
+            *rendered_texture_id = egui_ctx.add_image(images.bevy_icon.clone_weak());
         };
-        egui_ctx.set_egui_texture(BEVY_TEXTURE_ID, texture_handle);
     }
     if remove {
-        egui_ctx.remove_egui_texture(BEVY_TEXTURE_ID);
+        egui_ctx.remove_image(&images.bevy_icon);
+        egui_ctx.remove_image(&images.bevy_icon_inverted);
     }
 }
 
