@@ -27,6 +27,8 @@
 //!     App::new()
 //!         .add_plugins(DefaultPlugins)
 //!         .add_plugin(EguiPlugin)
+//!         // Systems that create Egui widgets should be run during the `CoreStage::Update` stage,
+//!         // or after the `EguiSystem::BeginFrame` system (which belongs to the `CoreStage::PreUpdate` stage).
 //!         .add_system(ui_example)
 //!         .run();
 //! }
@@ -47,7 +49,6 @@
 //! ## See also
 //!
 //! - [`bevy-inspector-egui`](https://github.com/jakobhellermann/bevy-inspector-egui)
-//! - [`bevy_megaui`](https://github.com/mvlabat/bevy_megaui)
 
 pub use egui;
 
@@ -59,22 +60,24 @@ use egui_node::EguiNode;
 use render_systems::EguiTransforms;
 
 use crate::systems::*;
+#[cfg(all(feature = "manage_clipboard", not(target_arch = "wasm32")))]
+use arboard::Clipboard;
 use bevy::{
     app::{App, CoreStage, Plugin, StartupStage},
-    asset::Handle,
-    ecs::schedule::{ParallelSystemDescriptorCoercion, SystemLabel},
+    asset::{AssetEvent, Assets, Handle, HandleId},
+    ecs::{
+        event::EventReader,
+        schedule::{ParallelSystemDescriptorCoercion, SystemLabel},
+        system::ResMut,
+    },
     input::InputSystem,
     log,
-    prelude::{AssetEvent, Assets, Commands, EventReader, ResMut},
     render::{render_graph::RenderGraph, texture::Image, RenderApp, RenderStage},
     utils::HashMap,
     window::WindowId,
 };
 #[cfg(all(feature = "manage_clipboard", not(target_arch = "wasm32")))]
-use clipboard::{ClipboardContext, ClipboardProvider};
-#[cfg(all(feature = "manage_clipboard", not(target_arch = "wasm32")))]
 use std::cell::{RefCell, RefMut};
-use std::collections::hash_map::Entry;
 #[cfg(all(feature = "manage_clipboard", not(target_arch = "wasm32")))]
 use thread_local::ThreadLocal;
 
@@ -122,7 +125,7 @@ pub struct EguiInput {
 #[derive(Default)]
 pub struct EguiClipboard {
     #[cfg(not(target_arch = "wasm32"))]
-    clipboard: ThreadLocal<Option<RefCell<ClipboardContext>>>,
+    clipboard: ThreadLocal<Option<RefCell<Clipboard>>>,
     #[cfg(target_arch = "wasm32")]
     clipboard: String,
 }
@@ -142,7 +145,7 @@ impl EguiClipboard {
     #[cfg(not(target_arch = "wasm32"))]
     fn set_contents_impl(&self, contents: &str) {
         if let Some(mut clipboard) = self.get() {
-            if let Err(err) = clipboard.set_contents(contents.to_owned()) {
+            if let Err(err) = clipboard.set_text(contents.to_owned()) {
                 log::error!("Failed to set clipboard contents: {:?}", err);
             }
         }
@@ -156,7 +159,7 @@ impl EguiClipboard {
     #[cfg(not(target_arch = "wasm32"))]
     fn get_contents_impl(&self) -> Option<String> {
         if let Some(mut clipboard) = self.get() {
-            match clipboard.get_contents() {
+            match clipboard.get_text() {
                 Ok(contents) => return Some(contents),
                 Err(err) => log::info!("Failed to get clipboard contents: {:?}", err),
             }
@@ -171,10 +174,10 @@ impl EguiClipboard {
     }
 
     #[cfg(not(target_arch = "wasm32"))]
-    fn get(&self) -> Option<RefMut<ClipboardContext>> {
+    fn get(&self) -> Option<RefMut<Clipboard>> {
         self.clipboard
             .get_or(|| {
-                ClipboardContext::new()
+                Clipboard::new()
                     .map(RefCell::new)
                     .map_err(|err| {
                         log::info!("Failed to initialize clipboard: {:?}", err);
@@ -188,25 +191,29 @@ impl EguiClipboard {
 
 /// Is used for storing Egui shapes. The actual resource is `HashMap<WindowId, EguiShapes>`.
 #[derive(Clone, Default, Debug)]
-pub struct EguiShapes {
+pub struct EguiRenderOutput {
     /// Pairs of rectangles and paint commands.
     ///
     /// The field gets populated during the [`EguiSystem::ProcessOutput`] system in the [`CoreStage::PostUpdate`] and reset during `EguiNode::update`.
     pub shapes: Vec<egui::epaint::ClippedShape>,
+
+    /// The change in egui textures since last frame.
+    pub textures_delta: egui::TexturesDelta,
 }
 
 /// Is used for storing Egui output. The actual resource is `HashMap<WindowId, EguiOutput>`.
 #[derive(Clone, Default)]
 pub struct EguiOutput {
-    /// The field gets updated during the [`EguiSystem::ProcessOutput`] system in the [`CoreStage::PostUpdate`]
-    pub output: egui::Output,
+    /// The field gets updated during the [`EguiSystem::ProcessOutput`] system in the [`CoreStage::PostUpdate`].
+    pub platform_output: egui::PlatformOutput,
 }
 
 /// A resource for storing `bevy_egui` context.
 #[derive(Clone)]
 pub struct EguiContext {
-    ctx: HashMap<WindowId, egui::CtxRef>,
-    egui_textures: HashMap<u64, Handle<Image>>,
+    ctx: HashMap<WindowId, egui::Context>,
+    user_textures: HashMap<HandleId, u64>,
+    last_texture_id: u64,
     mouse_position: Option<(WindowId, egui::Vec2)>,
 }
 
@@ -214,7 +221,8 @@ impl EguiContext {
     fn new() -> Self {
         Self {
             ctx: HashMap::default(),
-            egui_textures: Default::default(),
+            user_textures: Default::default(),
+            last_texture_id: 0,
             mouse_position: None,
         }
     }
@@ -225,8 +233,8 @@ impl EguiContext {
     /// The preferable way is to use `ctx_mut` to avoid unpredictable blocking inside UI systems.
     #[cfg(feature = "multi_threaded")]
     #[track_caller]
-    pub fn ctx(&self) -> &egui::CtxRef {
-        self.ctx.get(&WindowId::primary()).expect("`EguiContext::ctx` was called for an uninitialized context (primary window), consider moving your startup system to `StartupStage::Startup` stage or run it after `EguiStartupSystem::InitContexts` system")
+    pub fn ctx(&self) -> &egui::Context {
+        self.ctx.get(&WindowId::primary()).expect("`EguiContext::ctx` was called for an uninitialized context (primary window), consider moving your startup system to the `StartupStage::Startup` stage or run it after the `EguiStartupSystem::InitContexts` system")
     }
 
     /// Egui context for a specific window.
@@ -238,10 +246,10 @@ impl EguiContext {
     /// systems.
     #[cfg(feature = "multi_threaded")]
     #[track_caller]
-    pub fn ctx_for_window(&self, window: WindowId) -> &egui::CtxRef {
+    pub fn ctx_for_window(&self, window: WindowId) -> &egui::Context {
         self.ctx
             .get(&window)
-            .unwrap_or_else(|| panic!("`EguiContext::ctx_for_window` was called for an uninitialized context (window {}), consider moving your UI system to `CoreStage::Update` or run it after `EguiSystem::BeginFrame` system (`StartupStage::Startup` or `EguiStartupSystem::InitContexts` for startup systems respectively)", window))
+            .unwrap_or_else(|| panic!("`EguiContext::ctx_for_window` was called for an uninitialized context (window {}), consider moving your UI system to the `CoreStage::Update` stage or run it after the `EguiSystem::BeginFrame` system (`StartupStage::Startup` or `EguiStartupSystem::InitContexts` for startup systems respectively)", window))
     }
 
     /// Fallible variant of [`EguiContext::ctx_for_window`]. Make sure to set up the render graph by
@@ -251,29 +259,29 @@ impl EguiContext {
     /// The preferable way is to use `try_ctx_for_window_mut` to avoid unpredictable blocking inside
     /// UI systems.
     #[cfg(feature = "multi_threaded")]
-    pub fn try_ctx_for_window(&self, window: WindowId) -> Option<&egui::CtxRef> {
+    pub fn try_ctx_for_window(&self, window: WindowId) -> Option<&egui::Context> {
         self.ctx.get(&window)
     }
 
     /// Egui context of the primary window.
     #[track_caller]
-    pub fn ctx_mut(&mut self) -> &egui::CtxRef {
-        self.ctx.get(&WindowId::primary()).expect("`EguiContext::ctx_mut` was called for an uninitialized context (primary window), consider moving your startup system to `StartupStage::Startup` stage or run it after `EguiStartupSystem::InitContexts` system")
+    pub fn ctx_mut(&mut self) -> &egui::Context {
+        self.ctx.get(&WindowId::primary()).expect("`EguiContext::ctx_mut` was called for an uninitialized context (primary window), consider moving your startup system to the `StartupStage::Startup` stage or run it after the `EguiStartupSystem::InitContexts` system")
     }
 
     /// Egui context for a specific window.
     /// If you want to display UI on a non-primary window, make sure to set up the render graph by
     /// calling [`setup_pipeline`].
     #[track_caller]
-    pub fn ctx_for_window_mut(&mut self, window: WindowId) -> &egui::CtxRef {
+    pub fn ctx_for_window_mut(&mut self, window: WindowId) -> &egui::Context {
         self.ctx
             .get(&window)
-            .unwrap_or_else(|| panic!("`EguiContext::ctx_for_window_mut` was called for an uninitialized context (window {}), consider moving your UI system to `CoreStage::Update` or run it after `EguiSystem::BeginFrame` system (`StartupStage::Startup` or `EguiStartupSystem::InitContexts` for startup systems respectively)", window))
+            .unwrap_or_else(|| panic!("`EguiContext::ctx_for_window_mut` was called for an uninitialized context (window {}), consider moving your UI system to the `CoreStage::Update` stage or run it after the `EguiSystem::BeginFrame` system (`StartupStage::Startup` or `EguiStartupSystem::InitContexts` for startup systems respectively)", window))
     }
 
     /// Fallible variant of [`EguiContext::ctx_for_window_mut`]. Make sure to set up the render
     /// graph by calling [`setup_pipeline`].
-    pub fn try_ctx_for_window_mut(&mut self, window: WindowId) -> Option<&egui::CtxRef> {
+    pub fn try_ctx_for_window_mut(&mut self, window: WindowId) -> Option<&egui::Context> {
         self.ctx.get(&window)
     }
 
@@ -286,14 +294,14 @@ impl EguiContext {
     pub fn ctx_for_windows_mut<const N: usize>(
         &mut self,
         ids: [WindowId; N],
-    ) -> [&egui::CtxRef; N] {
-        let mut unique_ids = std::collections::HashSet::new();
+    ) -> [&egui::Context; N] {
+        let mut unique_ids = bevy::utils::HashSet::default();
         assert!(
             ids.iter().all(move |id| unique_ids.insert(id)),
             "Window ids passed to `EguiContext::ctx_for_windows_mut` must be unique: {:?}",
             ids
         );
-        ids.map(|id| self.ctx.get(&id).unwrap_or_else(|| panic!("`EguiContext::ctx_for_windows_mut` was called for an uninitialized context (window {}), consider moving your UI system to `CoreStage::Update` or run it after `EguiSystem::BeginFrame` system (`StartupStage::Startup` or `EguiStartupSystem::InitContexts` for startup systems respectively)", id)))
+        ids.map(|id| self.ctx.get(&id).unwrap_or_else(|| panic!("`EguiContext::ctx_for_windows_mut` was called for an uninitialized context (window {}), consider moving your UI system to the `CoreStage::Update` stage or run it after the `EguiSystem::BeginFrame` system (`StartupStage::Startup` or `EguiStartupSystem::InitContexts` for startup systems respectively)", id)))
     }
 
     /// Fallible variant of [`EguiContext::ctx_for_windows_mut`]. Make sure to set up the render
@@ -305,8 +313,8 @@ impl EguiContext {
     pub fn try_ctx_for_windows_mut<const N: usize>(
         &mut self,
         ids: [WindowId; N],
-    ) -> [Option<&egui::CtxRef>; N] {
-        let mut unique_ids = std::collections::HashSet::new();
+    ) -> [Option<&egui::Context>; N] {
+        let mut unique_ids = bevy::utils::HashSet::default();
         assert!(
             ids.iter().all(move |id| unique_ids.insert(id)),
             "Window ids passed to `EguiContext::ctx_for_windows_mut` must be unique: {:?}",
@@ -318,31 +326,25 @@ impl EguiContext {
     /// Can accept either a strong or a weak handle.
     ///
     /// You may want to pass a weak handle if you control removing texture assets in your
-    /// application manually and you don't want to bother with cleaning up textures in egui.
+    /// application manually and you don't want to bother with cleaning up textures in Egui.
     ///
-    /// You'll want to pass a strong handle if a texture is used only in egui and there's no
+    /// You'll want to pass a strong handle if a texture is used only in Egui and there's no
     /// handle copies stored anywhere else.
-    pub fn set_egui_texture(&mut self, id: u64, texture: Handle<Image>) {
-        log::debug!("Set egui texture: {:?}", texture);
-        self.egui_textures.insert(id, texture);
+    pub fn add_image(&mut self, image: Handle<Image>) -> egui::TextureId {
+        let id = *self.user_textures.entry(image.id).or_insert_with(|| {
+            let id = self.last_texture_id;
+            log::debug!("Add a new image (id: {}, handle: {:?})", id, image);
+            self.last_texture_id += 1;
+            id
+        });
+        egui::TextureId::User(id)
     }
 
-    /// Removes a texture handle associated with the id.
-    pub fn remove_egui_texture(&mut self, id: u64) {
-        let texture_handle = self.egui_textures.remove(&id);
-        log::debug!("Remove egui texture: {:?}", texture_handle);
-    }
-
-    // Is called when we get an event that a texture asset is removed.
-    #[allow(unused)] // TODO use
-    fn remove_texture(&mut self, texture_handle: &Handle<Image>) {
-        log::debug!("Removing egui handles: {:?}", texture_handle);
-        self.egui_textures = self
-            .egui_textures
-            .iter()
-            .map(|(id, texture)| (*id, texture.clone()))
-            .filter(|(_, texture)| texture != texture_handle)
-            .collect();
+    /// Removes the image handle and an Egui texture id associated with it.
+    pub fn remove_image(&mut self, image: &Handle<Image>) -> Option<u64> {
+        let id = self.user_textures.remove(&image.id);
+        log::debug!("Remove image (id: {:?}, handle: {:?})", id, image);
+        id
     }
 }
 
@@ -404,6 +406,17 @@ pub enum EguiSystem {
 
 impl Plugin for EguiPlugin {
     fn build(&self, app: &mut App) {
+        let world = &mut app.world;
+        world.insert_resource(EguiSettings::default());
+        world.insert_resource(HashMap::<WindowId, EguiInput>::default());
+        world.insert_resource(HashMap::<WindowId, EguiOutput>::default());
+        world.insert_resource(HashMap::<WindowId, WindowSize>::default());
+        world.insert_resource(HashMap::<WindowId, EguiRenderOutput>::default());
+        world.insert_resource(EguiManagedTextures::default());
+        #[cfg(feature = "manage_clipboard")]
+        world.insert_resource(EguiClipboard::default());
+        world.insert_resource(EguiContext::new());
+
         app.add_startup_system_to_stage(
             StartupStage::PreStartup,
             init_contexts_on_startup.label(EguiStartupSystem::InitContexts),
@@ -425,18 +438,11 @@ impl Plugin for EguiPlugin {
             CoreStage::PostUpdate,
             process_output.label(EguiSystem::ProcessOutput),
         );
-        app.add_system_to_stage(CoreStage::PostUpdate, update_egui_textures);
-
-        let world = &mut app.world;
-        world.get_resource_or_insert_with(EguiSettings::default);
-        world.get_resource_or_insert_with(HashMap::<WindowId, EguiInput>::default);
-        world.get_resource_or_insert_with(HashMap::<WindowId, EguiOutput>::default);
-        world.get_resource_or_insert_with(HashMap::<WindowId, WindowSize>::default);
-        world.get_resource_or_insert_with(HashMap::<WindowId, EguiShapes>::default);
-        world.get_resource_or_insert_with(EguiFontTextures::default);
-        #[cfg(feature = "manage_clipboard")]
-        world.get_resource_or_insert_with(EguiClipboard::default);
-        world.insert_resource(EguiContext::new());
+        app.add_system_to_stage(
+            CoreStage::PostUpdate,
+            update_egui_textures.after(EguiSystem::ProcessOutput),
+        );
+        app.add_system_to_stage(CoreStage::Last, free_egui_textures);
 
         if let Ok(render_app) = app.get_sub_app_mut(RenderApp) {
             render_app.init_resource::<egui_node::EguiPipeline>();
@@ -461,42 +467,97 @@ impl Plugin for EguiPlugin {
 }
 
 #[derive(Default)]
-pub(crate) struct EguiFontTextures(HashMap<WindowId, (Handle<Image>, u64)>);
+pub(crate) struct EguiManagedTextures(HashMap<(WindowId, u64), EguiManagedTexture>);
+
+pub(crate) struct EguiManagedTexture {
+    handle: Handle<Image>,
+    /// Stored in full so we can do partial updates (which bevy doesn't support).
+    color_image: egui::ColorImage,
+}
 
 fn update_egui_textures(
-    _commands: Commands,
+    mut egui_render_output: ResMut<HashMap<WindowId, EguiRenderOutput>>,
+    mut egui_managed_textures: ResMut<EguiManagedTextures>,
+    mut image_assets: ResMut<Assets<Image>>,
+) {
+    for (&window_id, egui_render_output) in egui_render_output.iter_mut() {
+        let set_textures = std::mem::take(&mut egui_render_output.textures_delta.set);
+
+        for (texture_id, image_delta) in set_textures {
+            let color_image = egui_node::as_color_image(&image_delta.image);
+
+            let texture_id = match texture_id {
+                egui::TextureId::Managed(texture_id) => texture_id,
+                egui::TextureId::User(_) => continue,
+            };
+
+            if let Some(pos) = image_delta.pos {
+                // Partial update.
+                if let Some(managed_texture) =
+                    egui_managed_textures.0.get_mut(&(window_id, texture_id))
+                {
+                    // TODO: when bevy supports it, only update the part of the texture that changes.
+                    update_image_rect(&mut managed_texture.color_image, pos, &color_image);
+                    let image = egui_node::color_image_as_bevy_image(&managed_texture.color_image);
+                    managed_texture.handle = image_assets.add(image);
+                } else {
+                    log::warn!("Partial update of a missing texture (id: {:?})", texture_id);
+                }
+            } else {
+                // Full update.
+                let image = egui_node::color_image_as_bevy_image(&color_image);
+                let handle = image_assets.add(image);
+                egui_managed_textures.0.insert(
+                    (window_id, texture_id),
+                    EguiManagedTexture {
+                        handle,
+                        color_image,
+                    },
+                );
+            }
+        }
+    }
+
+    fn update_image_rect(dest: &mut egui::ColorImage, [x, y]: [usize; 2], src: &egui::ColorImage) {
+        for sy in 0..src.height() {
+            for sx in 0..src.width() {
+                dest[(x + sx, y + sy)] = src[(sx, sy)];
+            }
+        }
+    }
+}
+
+fn free_egui_textures(
     mut egui_context: ResMut<EguiContext>,
-    mut egui_font_textures: ResMut<EguiFontTextures>,
+    mut egui_render_output: ResMut<HashMap<WindowId, EguiRenderOutput>>,
+    mut egui_managed_textures: ResMut<EguiManagedTextures>,
     mut image_assets: ResMut<Assets<Image>>,
     mut image_events: EventReader<AssetEvent<Image>>,
 ) {
-    egui_context.ctx.iter().for_each(|(&window_id, ctx)| {
-        let texture = ctx.font_image();
-
-        match egui_font_textures.0.entry(window_id) {
-            Entry::Occupied(entry) if entry.get().1 == texture.version => {}
-            Entry::Occupied(mut entry) => {
-                let image = image_assets.add(egui_node::as_wgpu_image(&texture));
-                entry.insert((image, texture.version));
+    for (&window_id, egui_render_output) in egui_render_output.iter_mut() {
+        let free_textures = std::mem::take(&mut egui_render_output.textures_delta.free);
+        for texture_id in free_textures {
+            if let egui::TextureId::Managed(texture_id) = texture_id {
+                let managed_texture = egui_managed_textures.0.remove(&(window_id, texture_id));
+                if let Some(managed_texture) = managed_texture {
+                    image_assets.remove(managed_texture.handle);
+                }
             }
-            Entry::Vacant(entry) => {
-                let image = image_assets.add(egui_node::as_wgpu_image(&texture));
-                entry.insert((image, texture.version));
-            }
-        };
-    });
+        }
+    }
 
     for image_event in image_events.iter() {
         if let AssetEvent::Removed { handle } = image_event {
-            egui_context.remove_texture(handle);
+            egui_context.remove_image(handle);
         }
     }
 }
 
 /// Egui's render graph config.
-#[allow(missing_docs)]
 pub struct RenderGraphConfig {
+    /// Target window.
     pub window_id: WindowId,
+    /// Render pass name.
     pub egui_pass: &'static str,
 }
 
@@ -513,11 +574,7 @@ impl Default for RenderGraphConfig {
 ///
 /// The pipeline for the primary window will already be set up by the [`EguiPlugin`],
 /// so you'll only need to manually call this if you want to use multiple windows.
-pub fn setup_pipeline(
-    render_graph: &mut RenderGraph,
-    config: RenderGraphConfig,
-    //msaa: &Msaa,
-) {
+pub fn setup_pipeline(render_graph: &mut RenderGraph, config: RenderGraphConfig) {
     render_graph.add_node(config.egui_pass, EguiNode::new(config.window_id));
 
     render_graph
@@ -565,7 +622,9 @@ mod tests {
     }
 
     #[test]
-    #[should_panic(expected = "Window ids passed to `EguiContext::ctx_for_windows_mut` must be unique")]
+    #[should_panic(
+        expected = "Window ids passed to `EguiContext::ctx_for_windows_mut` must be unique"
+    )]
     fn test_ctx_for_windows_mut_unique_check_panics() {
         let mut egui_context = EguiContext::new();
         let primary_window = WindowId::primary();
@@ -588,7 +647,9 @@ mod tests {
     }
 
     #[test]
-    #[should_panic(expected = "Window ids passed to `EguiContext::ctx_for_windows_mut` must be unique")]
+    #[should_panic(
+        expected = "Window ids passed to `EguiContext::ctx_for_windows_mut` must be unique"
+    )]
     fn test_try_ctx_for_windows_mut_unique_check_panics() {
         let mut egui_context = EguiContext::new();
         let primary_window = WindowId::primary();
